// Pedestal
PORTAL.Pedestal = object_add();
object_set_sprite(PORTAL.Pedestal, PORTAL.PropsS);

object_event_add(PORTAL.Pedestal, ev_create, 0, '
    // Align with mappixels.
    y += 2;

    image_speed = 0;
    color = make_color_rgb(255, 127, 0);
    defaultColor = color;
    hasPortalgun = true;
    randomColor = false;
    respawnTime = 0;

    ownedByMap = true; // Was this pedestal placed by the map?
    ds_list_add(PORTAL.pedestals, id);
');

object_event_add(PORTAL.Pedestal, ev_alarm, 0, '
    hasPortalgun = true;

    if (randomColor == true) {
        color = make_color_rgb(irandom(255), irandom(255), irandom(255));
    }
');

object_event_add(PORTAL.Pedestal, ev_destroy, 0, '
    ds_list_delete(PORTAL.pedestals, ds_list_find_index(PORTAL.pedestals, id));
');

object_event_add(PORTAL.Pedestal, ev_other, ev_user1, '
    color = readProperty(other.properties, "color", HEX, color);
    color = make_color_rgb((color&$ff0000)>>16, (color&$00ff00)>>8, color&$0000ff);

    respawnTime = readProperty(other.properties, "respawnTime", REAL, respawnTime);
    randomColor = readProperty(other.properties, "randomColor", BOOL, randomColor);

    if (randomColor == true) {
        color = make_color_rgb(irandom(255), irandom(255), irandom(255));
    }
    defaultColor = color;
');

object_event_add(PORTAL.Pedestal, ev_draw, 0, '
    draw_sprite_ext(sprite_index, 0, x, y, image_xscale, image_yscale, image_angle, c_white, image_alpha);

    if (hasPortalgun == true) {
        draw_sprite_ext(PORTAL.PortalgunS, 0, x - 10 * image_xscale, y + 6, image_xscale, image_yscale, image_angle, c_white, image_alpha);
    }

    if (distance_to_point(view_xview + view_wview / 2, view_yview + view_hview / 2) < 20) {
        var text;
        text = "Press " + chr(39) + chr(PORTAL.interactKey) + chr(39);

        draw_set_alpha(1);
        draw_set_valign(fa_bottom);
        draw_set_halign(fa_center);

        draw_set_color(c_black);
        draw_text(x, y - 20, text);

        draw_set_color(c_white);
        draw_text(x - 2, y - 22, text);
    }
');

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PORTAL.Portal = object_add();
object_set_sprite(PORTAL.Portal, PORTAL.PortalS);

object_event_add(PORTAL.Portal, ev_create, 0, '
    pair = noone;

    image_speed = 0;
    color = c_white;
    animationTime = 5 / global.delta_factor;

    ds_list_add(PORTAL.portals, id);
');

object_event_add(PORTAL.Portal, ev_alarm, 0, '
    if (instance_exists(pair) == true) {
        image_index = 1;
    }
');
object_event_add(PORTAL.Portal, ev_alarm, 1, '
    instance_destroy();
');

object_event_add(PORTAL.Portal, ev_destroy, 0, '
    ds_list_delete(PORTAL.portals, ds_list_find_index(PORTAL.portals, id));
    execute_string(PORTAL.SpawnParticles, x, y, image_angle, color);
');

object_event_add(PORTAL.Portal, ev_draw, 0, '
    var scale;
    scale = 1;
    if (alarm[0] > 0) {
        image_yscale = 1 - (alarm[0] / animationTime) * 0.75;
    } else if (alarm[1] > 0) {
        image_yscale = 0.25 + (alarm[1] / animationTime) * 0.75;
    }

    draw_sprite_ext(sprite_index, image_index, x, y, image_xscale, image_yscale, image_angle, color, image_alpha);
');

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PORTAL.Portalgun = object_add();
object_set_parent(PORTAL.Portalgun, Blade); // Blade is still pretty much the best choice.
object_set_sprite(PORTAL.Portalgun, PORTAL.PortalgunS);
object_set_depth(PORTAL.Portalgun, -1);

object_event_add(PORTAL.Portalgun, ev_create, 0, '
    event_inherited();

    xoffset = 0;
    yoffset = 2;

    refireTime = 40;
    colorR = 255;
    colorG = 127;
    colorB = 0;
    lastShotColor = 0;

    primary = noone;
    secondary = noone;
    holding = noone;

    normalSprite = PORTAL.PortalgunS;
    recoilSprite = PORTAL.PortalgunS;
    reloadSprite = -1;

    recoilTime = 1;
    recoilAnimLength = sprite_get_number(recoilSprite) / 2;
    recoilImageSpeed = recoilAnimLength/recoilTime;

    // Prevent crashes...
    t = 0;
    unscopedDamage = 0;
    baseDamage = 0;
    uberCharge = 0;
    lobbed = 0;
    readyToStab = true; // True or Spy cant uncloak!
    readyToBlast = false;
    readyToFlare = false;
');

object_event_add(PORTAL.Portalgun, ev_step, ev_step_end, '
    event_inherited();

    // Do not want.
    if (ownerPlayer == global.myself) {
        with(AmmoCounter) {
            instance_destroy();
        }

        with(StickyCounter) {
            instance_destroy();
        }
    }

    if (holding != noone) {
        if (holding.object_index != PORTAL.Cube && holding.held == true) {
            holding.alarm[10] = 2;
            holding.x = round(x + xoffset * image_xscale + lengthdir_x(45, owner.aimDirection));
            holding.y = round(y + yoffset + lengthdir_y(45, owner.aimDirection)) + owner.equipmentOffset;
        }
    }
');

object_event_add(PORTAL.Portalgun, ev_destroy, 0, '
    if (holding != noone) {
        with(holding) {
            if (held == true) {
                image_alpha = 1;
                x = round(other.x + other.xoffset * other.image_xscale + lengthdir_x(45, other.owner.aimDirection));
                y = round(other.y + other.yoffset + lengthdir_y(45, other.owner.aimDirection)) + other.owner.equipmentOffset;

                event_user(3);
            }
        }
    }
');

object_event_add(PORTAL.Portalgun, ev_other, ev_user1, '
    if (readyToShoot && !owner.cloak && !instance_exists(holding)) {
       playsound(x, y, PORTAL.ShotSnd);

        var shot;
        randomize();
        shot = createShot(x + lengthdir_x(8, owner.aimDirection), y + lengthdir_y(8, owner.aimDirection), PORTAL.Projectile, DAMAGE_SOURCE_PORTAL, owner.aimDirection, 17);
        shot.gun = id;
        shot.color = make_color_rgb(colorR, colorG, colorB);
        shot.primary = true;

        justShot = true;
        readyToShoot = false;
        alarm[0] = refireTime / global.delta_factor;
        lastShotColor = shot.color;
    }
');

object_event_add(PORTAL.Portalgun, ev_other, ev_user2, '
    if (readyToShoot && !owner.cloak && !instance_exists(holding)) {
        playsound(x, y, PORTAL.ShotSnd);

        var shot;
        randomize();
        shot = createShot(x + lengthdir_x(8, owner.aimDirection), y + lengthdir_y(8, owner.aimDirection), PORTAL.Projectile, DAMAGE_SOURCE_PORTAL, owner.aimDirection, 17);
        shot.gun = id;
        shot.color = make_color_rgb(255 - colorR, 255 - colorG, 255 - colorB);
        shot.primary = false;

        justShot = true;
        readyToShoot = false;
        alarm[0] = refireTime / global.delta_factor;
        lastShotColor = shot.color;
    }
');

// Simulate the original weapons to prevent clients from locking up.
object_event_add(PORTAL.Portalgun, ev_other, ev_user12, "
    switch(ownerPlayer.class) {
        case CLASS_MEDIC:
            write_ubyte(global.serializeBuffer, 255);
        break;
        case CLASS_DEMOMAN:
            event_inherited();
            write_ubyte(global.serializeBuffer, 0);
        break;
        default:
            event_inherited();
        break;
    }
");

object_event_add(PORTAL.Portalgun, ev_other, ev_user14, '
    if (global.isHost == true) {
        with(primary) {
            execute_string(PORTAL.SendDestroyPortal, id);
        }

        with(secondary) {
            execute_string(PORTAL.SendDestroyPortal, id);
        }

        with(holding) {
            if (held == true) {
                x = round(other.x + other.xoffset * image_xscale + lengthdir_x(45, other.owner.aimDirection));
                y = round(other.y + other.yoffset + lengthdir_y(45, other.owner.aimDirection)) + other.owner.equipmentOffset;
                event_user(3);
            }
        }
        holding = noone;
    }
');

object_event_add(PORTAL.Portalgun, ev_draw, 0, '
    if (distance_to_point(view_xview + view_wview/2, view_yview + view_hview/2) > 800) {
        exit;
    }

    if (!owner.invisible and !owner.taunting and !owner.omnomnomnom and !owner.player.humiliated) {
        if (!owner.cloak) {
            image_alpha = power(owner.cloakAlpha, 0.5);
        } else {
            image_alpha = power(owner.cloakAlpha, 2);
        }

        draw_sprite_ext(sprite_index, 0, round(x+xoffset*image_xscale), round(y+yoffset) + owner.equipmentOffset, image_xscale, image_yscale, image_angle, c_white, image_alpha);
        if (holding != noone) {
            if (holding.object_index == PORTAL.Cube) {
                draw_sprite_ext(holding.sprite_index, holding.image_index, round(x + xoffset * image_xscale + lengthdir_x(45, owner.aimDirection)), round(y + yoffset+ lengthdir_y(45, owner.aimDirection)) + owner.equipmentOffset, holding.image_xscale, holding.image_yscale, 0, c_white, image_alpha);
            }
        }

        if (owner.ubered) {
            if (owner.team == TEAM_RED) {
                ubercolour = c_red;
            } else if (owner.team == TEAM_BLUE) {
                ubercolour = c_blue;
            }

            draw_sprite_ext(sprite_index, 0, round(x+xoffset*image_xscale), round(y+yoffset) + owner.equipmentOffset, image_xscale, image_yscale, image_angle, ubercolour, 0.7*image_alpha);
        }

        if (lastShotColor != 0) {
            draw_sprite_ext(PORTAL.PortalgunParticlesS, 0, round(x+xoffset*image_xscale), round(y+yoffset) + owner.equipmentOffset, image_xscale, image_yscale, image_angle, lastShotColor, image_alpha);
        }
    }

    if (instance_exists(primary) == true) {
        draw_sprite_ext(PORTAL.PortalCrosshairS, 0, mouse_x, mouse_y, 1, 1, 0, primary.color, 1);
    }

    if (instance_exists(secondary) == true) {
        draw_sprite_ext(PORTAL.PortalCrosshairS, 1, mouse_x, mouse_y, 1, 1, 0, secondary.color, 1);
    }
');

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PORTAL.Cube = object_add();
object_set_parent(PORTAL.Cube, DropdownPlatform);
object_set_sprite(PORTAL.Cube, PORTAL.CubeS);

object_event_add(PORTAL.Cube, ev_create, 0, '
    ds_list_add(PORTAL.cubes, id);

    sprite_index = PORTAL.CubeS;
    image_speed = 0;
    target = noone;
    synced = false;
    held = false;
    lastPortal = noone;
    pushDir = 270;
');

object_event_add(PORTAL.Cube, ev_alarm, 0, '
    image_index = 2;
');

object_event_add(PORTAL.Cube, ev_alarm, 2, '
    pushDir = 270;
');

object_event_add(PORTAL.Cube, ev_step, ev_step_normal, '
    if (image_alpha == 0) {
        exit;
    }

    if (lastPortal != noone) {
        if (place_meeting(x, y, lastPortal) == false) {
            lastPortal = noone;
        }
    }

    // We perform the physics in the normal step so that the excursion funnel can set the speed in begin step
    // and then we solve collisions after.
    charSetAllSolids();
    with(Obstacle) {
        solid = true;
    }
    with(DropdownPlatform) {
        solid = true;
    }

    if (place_free(x, y+1)) {
        vspeed += 0.6 * global.delta_factor;
        if (vspeed > 10) {
            vspeed = 10;
        }

        hspeed = hspeed * delta_mult(0.96);
    } else {
        hspeed = hspeed * delta_mult(0.8);
    }

    if(!place_free(x+hspeed, y+vspeed)) {
        event_user(2);
    } else {
        x += hspeed * global.delta_factor;
        y += vspeed * global.delta_factor;
    }

    if (global.isHost == true && abs(hspeed) < 0.1 && abs(vspeed) < 0.1 && synced == false) {
        execute_string(PORTAL.SendSyncCube, id);
    }

    if (hspeed != 0 || vspeed != 0) {
        synced = false;
        if (target != noone) {
            target.pressed = false;
            target = noone;
        }
    }

    x -= hspeed;
    y -= vspeed;

    charUnsetSolids();
    with(DropdownPlatform) {
        solid = false;
    }
');

object_event_add(PORTAL.Cube, ev_draw, 0, '
    draw_sprite_ext(sprite_index, image_index, x, y, image_xscale, image_yscale, image_angle, c_white, image_alpha);
');

object_event_add(PORTAL.Cube, ev_destroy, 0, '
    ds_list_delete(PORTAL.cubes, ds_list_find_index(PORTAL.cubes, id));
');

object_event_add(PORTAL.Cube, ev_collision, PORTAL.Portal, '
    execute_string(PORTAL.TeleportProjectile, id, other.id, true);
');

object_event_add(PORTAL.Cube, ev_collision, KillBox, '
    event_user(3);
');

object_event_add(PORTAL.Cube, ev_collision, PitFall, '
    event_user(3);
');

object_event_add(PORTAL.Cube, ev_collision, FragBox, '
    event_user(3);
');

object_event_add(PORTAL.Cube, ev_other, ev_user1, '
    image_index = readProperty(other.properties, "form", REAL, image_index);
');

object_event_add(PORTAL.Cube, ev_other, ev_user2, '
    var oldx, oldy, hleft, vleft, bboxheight, bboxwidth;
    oldx=x;
    oldy=y;
    bboxheight = bbox_bottom-bbox_top;
    bboxwidth = bbox_right-bbox_left;

    // slide in an appropriate direction to get outside of walls
    if(!place_free(x, y))
    {
        var uy, dy, lx, rx, distu, distd, distl, distr;

        move_outside_solid(90, bboxheight/2);
        distu = oldy - y;
        uy = y;
        y = oldy;

        move_outside_solid(270, bboxheight/2);
        distd = y - oldy;
        dy = y;
        y = oldy;

        move_outside_solid(0, bboxwidth/2);
        distr = x - oldx;
        rx = x;
        x = oldx;

        move_outside_solid(180, bboxwidth/2);
        distl = oldx - x;
        lx = x;
        x = oldx;

        if(distu < distd and distu < distr and distu < distl)
            y = uy;
        else if(distd < distr and distd < distl)
            y = dy;
        else if(distr < distl)
            x = rx;
        else
            x = lx;

        if(!place_free(x, y))
        {
            x = oldx;
            y = oldy;
        }
    }

    hleft = hspeed * global.delta_factor;
    vleft = vspeed * global.delta_factor;

    var loopCounter, stuck;
    loopCounter = 0;
    stuck = 0;
    while((abs(hleft) > 0.1 || abs(vleft) > 0.1) && stuck = 0) { // while we still have distance to travel
        loopCounter += 1;
        if(loopCounter > 3) {
            stuck = 1;
        }

        var prevX, prevY, collisionRectified;
        collisionRectified = false; // set this to true when we fix a collision problem
        prevX = x;
        prevY = y;
        really_move_contact_solid(point_direction(x, y, x + hleft, y + vleft), point_distance(x, y, x + hleft, y + vleft));
        hleft -= x - prevX;
        vleft -= y - prevY;

        if(vleft != 0 && !place_free(x, y + sign(vleft))) { // we hit a ceiling or floor
            vleft = 0;
            vspeed = 0;
            collisionRectified = true;
        }
        if(hleft != 0 && !place_free(x + sign(hleft), y)) { // we hit a wall on the left or right
            hleft = 0;
            hspeed = 0;
            collisionRectified = true;
        }
        if(!collisionRectified && (abs(hleft) >= 1 || abs(vleft) >= 1)) {
            vspeed = 0;
            vleft = 0;
        }
    }
');

object_event_add(PORTAL.Cube, ev_other, ev_user3, '
    // Fizzle the shizzle!
    if (global.isHost == true) {
        image_alpha = 0;
        execute_string(PORTAL.SendSyncCube, id);
    }
');

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PORTAL.Button = object_add();
object_set_sprite(PORTAL.Button, PORTAL.PropsS);
object_set_parent(PORTAL.Button, PORTAL.Sender);

object_event_add(PORTAL.Button, ev_create, 0, '
    event_inherited();
    timer = 90;
    image_index = 1;
');

object_event_add(PORTAL.Button, ev_other, ev_user1, '
    event_inherited();
    timer = readProperty(other.properties, "timer", REAL, timer);
');

object_event_add(PORTAL.Button, ev_draw, 0, '
    draw_sprite_ext(sprite_index, pressed + 1, x, y, image_xscale, image_yscale, image_angle, c_white, image_alpha);

    if (alarm[0] > 0) {
        draw_intel_timer(0, alarm[0] / timer * global.delta_factor);
    }
');

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PORTAL.Plate = object_add();
object_set_sprite(PORTAL.Plate, PORTAL.PropsS);
object_set_parent(PORTAL.Plate, PORTAL.Sender);

object_event_add(PORTAL.Plate, ev_create, 0, '
    event_inherited();
    lastCube = noone;
    image_index = 3;
');

object_event_add(PORTAL.Plate, ev_alarm, 0, '
    pressed = false;

    if (lastCube != noone) {
        lastCube.target = collision_line(lastCube.x, lastCube.y, lastCube.x + lengthdir_x(15, lastCube.pushDir), lastCube.y + lengthdir_y(15, lastCube.pushDir), id, true, false);
        if (lastCube.target != noone) {
            pressed = true;
        }
    }
');

object_event_add(PORTAL.Plate, ev_draw, 0, '
    draw_sprite_ext(sprite_index, pressed + 3, x, y, image_xscale, image_yscale, image_angle, c_white, image_alpha);
');

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PORTAL.Decal = object_add();
object_set_sprite(PORTAL.Decal, PORTAL.DecalsS);
object_set_depth(PORTAL.Decal, 1);

object_event_add(PORTAL.Decal, ev_create, 0, '
    image_speed = 0;
    image_index = 0;

    decalId = 0;
    button = noone;
');

object_event_add(PORTAL.Decal, ev_other, ev_user1, '
    decalId = readProperty(other.properties, "decalId", REAL, decalId);
');

object_event_add(PORTAL.Decal, ev_draw, 0, '
    if (button == noone) {
        var i;
        for (i = 0; i < ds_list_size(PORTAL.buttons); i += 1) {
            button = ds_list_find_value(PORTAL.buttons, i);
            if (button.decalId == decalId) {
                break;
            }
        }
    }

    if (button != noone) {
        image_index = button.pressed;
        var x1, y1, x2, y2;
        x1 = round(button.x / 32) * 32;
        y1 = round(button.y / 32) * 32;
        x2 = round(x / 32) * 32;
        y2 = round(y / 32) * 32;

        while(y1 > y2) {
            draw_sprite(sprite_index, image_index + 2, x1, y1);
            y1 -= 32;
        }
        while(y1 < y2) {
            draw_sprite(sprite_index, image_index + 2, x1, y1);
            y1 += 32;
        }

        while(x1 > x2) {
            draw_sprite(sprite_index, image_index + 2, x1, y1);
            x1 -= 32;
        }
        while(x1 < x2) {
            draw_sprite(sprite_index, image_index + 2, x1, y1);
            x1 += 32;
        }
    }

    draw_sprite_ext(sprite_index, image_index, x, y, image_xscale, image_yscale, image_angle, c_white, image_alpha);
');

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PORTAL.Door = object_add();
object_set_sprite(PORTAL.Door, PORTAL.DoorS);
object_set_parent(PORTAL.Door, PORTAL.Receiver);

object_event_add(PORTAL.Door, ev_create, 0, '
    event_inherited();

    image_index = 0;
    gate = noone;
');

object_event_add(PORTAL.Door, ev_step, ev_step_normal, '
    event_inherited();

    if (gate == noone && active == false) {
        gate = instance_create(x - 7, y - 22, Obstacle);
        gate.image_yscale = 3;
    } else if (gate != noone && active == true) {
        with(gate) {
            instance_destroy();
        }
        gate = noone;
    }

    image_index = active;
');

object_event_add(PORTAL.Door, ev_draw, 0, '
    draw_sprite_ext(sprite_index, image_index, x, y, image_xscale, image_yscale, image_angle, c_white, image_alpha);
');


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PORTAL.Beam = object_add();
object_set_sprite(PORTAL.Beam, PORTAL.PropsS);
object_set_depth(PORTAL.Beam, -1);
object_set_parent(PORTAL.Beam, PORTAL.Receiver);

object_event_add(PORTAL.Beam, ev_create, 0, '
    event_inherited();
    dps = 90;
    points = ds_list_create();

    image_index = 9;
');

object_event_add(PORTAL.Beam, ev_step, ev_step_normal, '
    event_inherited();

    if (active == true) {
        event_user(2);
    }
');

object_event_add(PORTAL.Beam, ev_draw, 0, '
    draw_sprite_ext(PORTAL.PropsS, 9 + active, x, y, image_xscale, image_yscale, image_angle, c_white, image_alpha);

    if (active == true) {
        var i;
        for(i = 0; i < ds_list_size(points); i += 4 ) {
            draw_set_alpha(0.2);
            draw_line_width_color(ds_list_find_value(points, i), ds_list_find_value(points, i + 1), ds_list_find_value(points, i + 2), ds_list_find_value(points, i + 3), 2, c_red, c_red);
            draw_line_width_color(ds_list_find_value(points, i), ds_list_find_value(points, i + 1), ds_list_find_value(points, i + 2), ds_list_find_value(points, i + 3), 6, c_red, c_red);
            draw_set_alpha(1);
        }
    }
');

object_event_add(PORTAL.Beam, ev_destroy, 0, '
    event_inherited();
    ds_list_destroy(points);
');

object_event_add(PORTAL.Beam, ev_other, ev_user1, '
    event_inherited();
    dps = readProperty(other.properties, "dps", REAL, dps);
');

object_event_add(PORTAL.Beam, ev_other, ev_user2, '
    ds_list_clear(points);
    gunSetSolids();
    var i, originalX, leftmostX, x1, y1, dir, j, max, new;
    originalX = ds_map_create();
    max = 10;

    dir = image_angle;
    x1 = x - lengthdir_x(19, dir);
    y1 = y - lengthdir_y(19, dir);
    leftmostX = -8008135;

    for (i = 0; i < ds_list_size(PORTAL.BEAMCOLLIDERS); i += 1) {
        if (!solid) {
            ds_map_add(originalX, id, x);
            x += leftmostX - bbox_right - 10;
        }
    }

    for (j = 0; j < max; j += 1) {
        distance = execute_string(PORTAL.Raycast, PORTAL.BEAMCOLLIDERS, 1000, x1, y1, x1 + lengthdir_x(1000, dir), y1 + lengthdir_y(1000, dir));
        if (distance > 1) {
            ds_list_add(points, x1);
            ds_list_add(points, y1);
            ds_list_add(points, x1 + lengthdir_x(distance, dir));
            ds_list_add(points, y1 + lengthdir_y(distance, dir));
        }

        new = false;
        if (hitInstance != noone) {
            if (hitInstance.object_index == PORTAL.Cube) {
                if (hitInstance.image_index = 2 || hitInstance.image_index = 3) {
                    dir = (dir + 90) mod 360;
                    x1 = hitInstance.x + lengthdir_x(11, dir);
                    y1 = hitInstance.y + lengthdir_y(11, dir);

                    hitInstance.image_index = 3;
                    hitInstance.alarm[0] = 2;

                    with(hitInstance) {
                        ds_map_add(originalX, id, x);
                        x += leftmostX - bbox_right - 10;
                    }

                    new = true;
                }
            } else if (hitInstance.object_index == PORTAL.Portal) {
                if (instance_exists(hitInstance.pair) == true) {
                    dir = 180 + hitInstance.image_angle - dir + hitInstance.pair.image_angle;
                    x1 = hitInstance.pair.x;
                    y1 = hitInstance.pair.y;

                    // Remove the portal so we dont collide with it again.
                    with(hitInstance.pair) {
                        ds_map_add(originalX, id, x);
                        x += leftmostX - bbox_right - 10;
                    }

                    new = true;
                }
            } else if (hitInstance.object_index == PORTAL.Lens) {
                if (hitInstance.pressed == false) {
                    playsound(hitInstance.x, hitInstance.y, PORTAL.ShotSnd);
                }

                hitInstance.pressed = true;
                hitInstance.alarm[0] = 2;
            } else if (hitInstance.object_index == Character or object_is_ancestor(hitInstance.object_index, Character) == true) {
                hitInstance.lastDamageSource = DAMAGE_SOURCE_FINISHED_OFF_GIB;
                damageCharacter(PORTAL.MapPlayerInstance, hitInstance.id, dps * global.delta_factor / 30);

                if (hitInstance.lastDamageDealer != PORTAL.MapPlayerInstance)
                {
                    hitInstance.secondToLastDamageDealer = hitInstance.lastDamageDealer;
                    hitInstance.alarm[4] = hitInstance.alarm[3]
                }
                hitInstance.alarm[3] = ASSIST_TIME / global.delta_factor;
                hitInstance.lastDamageDealer = PORTAL.MapPlayerInstance;
                dealFlicker(hitInstance.id);
                if (global.gibLevel > 0 && random(1) > 0.9) {
                    blood = instance_create(hitInstance.x, hitInstance.y, Blood);
                    blood.direction = dir - 180;
                }
            } else if (hitInstance.object_index == Sentry or object_is_ancestor(hitInstance.object_index, Sentry) == true) {
                hitInstance.lastDamageSource = DAMAGE_SOURCE_FINISHED_OFF_GIB;
                damageSentry(PORTAL.MapPlayerInstance, hitInstance.id, dps * global.delta_factor / 30);
            } else if (hitInstance.object_index == Generator or object_is_ancestor(hitInstance.object_index, Generator) == true) {
                damageGenerator(PORTAL.MapPlayerInstance, hitInstance.id, dps * global.delta_factor / 30);
            }
        }

        // We hit nothing special.
        if (new == false) {
            break;
        }
    }

    var key, origX;
    for(key = ds_map_find_first(originalX); key; key = ds_map_find_next(originalX, key)) {
        origX = ds_map_find_value(originalX, key);
        with(key) {
            x = origX;
        }
    }

    ds_map_destroy(originalX);
    gunUnsetSolids();
');

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PORTAL.Lens = object_add();
object_set_sprite(PORTAL.Lens, PORTAL.PropsS);
object_set_parent(PORTAL.Lens, PORTAL.Sender);

object_event_add(PORTAL.Lens, ev_create, 0, '
    event_inherited();
    image_index = 11;
');

object_event_add(PORTAL.Lens, ev_draw, 0, '
    draw_sprite_ext(sprite_index, 11 + pressed, x, y, image_xscale, image_yscale, image_angle, c_white, image_alpha);
');

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PORTAL.ExcursionFunnel = object_add();
object_set_sprite(PORTAL.ExcursionFunnel, PORTAL.PropsS);
object_set_parent(PORTAL.ExcursionFunnel, PORTAL.Receiver);

object_event_add(PORTAL.ExcursionFunnel, ev_create, 0, '
    event_inherited();
    image_index = 5;

    canReverse = false;
    reverse = false;
    animIndex = 0;
    points = ds_list_create();
    update = true;
');

object_event_add(PORTAL.ExcursionFunnel, ev_step, ev_step_begin, '
    if (active == true) {
        if (update == true) {
            event_user(2);
            update = false;
        }

        var i, length, dir, _x, _y, image;
        for (i = 0; i < ds_list_size(points); i += 4) {
            _x = ds_list_find_value(points, i);
            _y = ds_list_find_value(points, i + 1);
            length = ds_list_find_value(points, i + 2);
            dir = ds_list_find_value(points, i + 3);

            // Move all cubes in the funnel.
            var j, x2, y2, cube, dist, percent;
            x2 = _x + lengthdir_x(length, dir);
            y2 = _y + lengthdir_y(length, dir);
            for (j = 0; j < ds_list_size(PORTAL.cubes); j += 1) {
                cube = ds_list_find_value(PORTAL.cubes, j);
                // Dont move cubes that are being held by a portalgun.
                if (cube.held == true) {
                    continue;
                }

                if (collision_line(_x, _y, x2, y2, cube, false, false) > 0) {
                    if (reverse == true) {
                        percent = median(0, point_distance(_x, _y, cube.x, cube.y) / length - 1 / length, 1);
                    } else {
                        percent = median(0, point_distance(_x, _y, cube.x, cube.y) / length + 1 / length, 1);
                    }

                    cube.hspeed = (_x + lengthdir_x(length * percent, dir)) - cube.x;
                    cube.vspeed = (_y + lengthdir_y(length * percent, dir)) - cube.y - 0.6;
                    cube.pushDir = dir;
                    cube.alarm[2] = 2;
                }
            }
        }
    }
');

object_event_add(PORTAL.ExcursionFunnel, ev_step, ev_step_normal, '
    event_inherited();
    if (canReverse == true) {
        reverse = active;
        active = true;
    }
');

object_event_add(PORTAL.ExcursionFunnel, ev_draw, 0, '
    animIndex = (animIndex + 0.2 * global.delta_factor) mod 6;

    if (active == true) {
        if (reverse == true) {
            image_index = 7;
        } else {
            image_index = 6;
        }
    } else {
        image_index = 5;
    }
    draw_sprite_ext(sprite_index, image_index, x, y, image_xscale, image_yscale, image_angle, c_white, image_alpha);

    if (active == true) {
        var i, length, dir, _x, _y, image;
        for (i = 0; i < ds_list_size(points); i += 4) {
            _x = ds_list_find_value(points, i);
            _y = ds_list_find_value(points, i + 1);
            length = ds_list_find_value(points, i + 2);
            dir = ds_list_find_value(points, i + 3);
            image = animIndex;

            while(length > 0) {
                if (length > 32) {
                    draw_sprite_ext(PORTAL.ExcursionFunnelS, floor(image + reverse * 7), _x, _y, 1, 1, dir, c_white, 0.8);
                    _x += lengthdir_x(32, dir);
                    _y += lengthdir_y(32, dir);
                    length -= 32;
                } else {
                    draw_sprite_general(PORTAL.ExcursionFunnelS, floor(image + reverse * 7), 0, 0, length, 32, _x + lengthdir_x(16, dir + 90), _y + lengthdir_y(16, dir + 90), 1, 1, dir, c_white, c_white, c_white, c_white, 0.8);
                    length = 0;
                }

                image = (image + 1) mod 6;
            }
        }
    }
');

object_event_add(PORTAL.ExcursionFunnel, ev_destroy, 0, '
    event_inherited();
    ds_list_destroy(points);
');

object_event_add(PORTAL.ExcursionFunnel, ev_other, ev_user1, '
    event_inherited();
    canReverse = readProperty(other.properties, "reverse", BOOL, canReverse);
');

object_event_add(PORTAL.ExcursionFunnel, ev_other, ev_user2, '
    ds_list_clear(points);
    gunSetSolids();
    var i, originalX, leftmostX, x1, y1, dir, j, max, new;
    originalX = ds_map_create();
    max = 3;

    dir = image_angle;
    x1 = x - lengthdir_x(23, dir);
    y1 = y - lengthdir_y(23, dir);
    leftmostX = -8008135;

    for (i = 0; i < ds_list_size(PORTAL.ROCKETCOLLIDERS); i += 1) {
        if (!solid) {
            ds_map_add(originalX, id, x);
            x += leftmostX - bbox_right - 10;
        }
    }

    for (j = 0; j < max; j += 1) {
        distance = execute_string(PORTAL.Raycast, PORTAL.ROCKETCOLLIDERS, 2000, x1, y1, x1 + lengthdir_x(2000, dir), y1 + lengthdir_y(2000, dir));
        if (distance > 1) {
            ds_list_add(points, x1);
            ds_list_add(points, y1);
            ds_list_add(points, distance);
            ds_list_add(points, dir);
        }

        new = false;
        if (hitInstance != noone) {
            if (hitInstance.object_index == PORTAL.Portal) {
                if (instance_exists(hitInstance.pair) == true) {
                    dir = (180 + hitInstance.image_angle - dir + hitInstance.pair.image_angle) mod 360;
                    x1 = hitInstance.pair.x;
                    y1 = hitInstance.pair.y;

                    // Remove the portal so we dont collide with it again.
                    with(hitInstance.pair) {
                        ds_map_add(originalX, id, x);
                        x += leftmostX - bbox_right - 10;
                    }

                    new = true;
                }
            }
        }

        if (new == false) {
            break;
        }
    }

    var key, origX;
    for(key = ds_map_find_first(originalX); key; key = ds_map_find_next(originalX, key)) {
        origX = ds_map_find_value(originalX, key);
        with(key) {
            x = origX;
        }
    }

    ds_map_destroy(originalX);
    gunUnsetSolids();
');

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PORTAL.LightBridge = object_add();
object_set_sprite(PORTAL.LightBridge, PORTAL.PropsS);
object_set_parent(PORTAL.LightBridge, PORTAL.Receiver);

object_event_add(PORTAL.LightBridge, ev_create, 0, '
    event_inherited();
    image_index = 8;

    points = ds_list_create();
    update = true;
    wasActive = false;

    // Just keep a reference to 3 walls and 3 dropdown platforms.
    bridgePart[0] = instance_create(-999, -999, DropdownPlatform);
    bridgePart[1] = instance_create(-999, -999, DropdownPlatform);
    bridgePart[2] = instance_create(-999, -999, DropdownPlatform);
    bridgePart[3] = instance_create(-999, -999, PlayerWallHorizontal);
    bridgePart[4] = instance_create(-999, -999, PlayerWallHorizontal);
    bridgePart[5] = instance_create(-999, -999, PlayerWallHorizontal);
');

object_event_add(PORTAL.LightBridge, ev_step, ev_step_normal, '
    event_inherited();

    if (active == true && update == true) {
        event_user(2);
        wasActive = !active;
        update = false;
    }

    if (active != wasActive) {
        var i, length, dir, _x, _y, bridge;
        for (i = 0; i < 6; i += 1) {
            bridgePart[i].x = -999;
            bridgePart[i].y = -999;
            bridgePart[i].image_xscale = 1;
            bridgePart[i].image_yscale = 1;
        }

        for (i = 0; i < ds_list_size(points); i += 4) {
            _x = ds_list_find_value(points, i);
            _y = ds_list_find_value(points, i + 1);
            length = ds_list_find_value(points, i + 2);
            dir = ds_list_find_value(points, i + 3);
            if (dir == 0 || dir == 180) {
                bridge = bridgePart[i / 4];
            } else {
                bridge = bridgePart[i / 4 + 3];
            }

            bridge.x = _x;
            bridge.y = _y;
            bridge.image_xscale = length / 60;
            bridge.image_angle = dir;
        }
    }
    wasActive = active;
');

object_event_add(PORTAL.LightBridge, ev_destroy, 0, '
    event_inherited();
    ds_list_destroy(points);
');

object_event_add(PORTAL.LightBridge, ev_draw, 0, '
    draw_sprite_ext(sprite_index, image_index, x, y, image_xscale, image_yscale, image_angle, c_white, image_alpha);

    if (active == true) {
        var i, length, dir, _x, _y;
        for (i = 0; i < ds_list_size(points); i += 4) {
            _x = ds_list_find_value(points, i);
            _y = ds_list_find_value(points, i + 1);
            length = ds_list_find_value(points, i + 2);
            dir = ds_list_find_value(points, i + 3);

            draw_set_alpha(0.6);
            draw_line_width_color(_x, _y, _x + lengthdir_x(length, dir), _y + lengthdir_y(length, dir), 4, c_blue, c_blue);
            draw_set_alpha(1);
        }
    }
');

object_event_add(PORTAL.LightBridge, ev_other, ev_user2, '
    ds_list_clear(points);
    gunSetSolids();
    var i, originalX, leftmostX, x1, y1, dir, j, max, new;
    originalX = ds_map_create();
    max = 3;

    dir = image_angle;
    x1 = x - lengthdir_x(23, dir);
    y1 = y - lengthdir_y(23, dir);
    leftmostX = -8008135;

    for (i = 0; i < ds_list_size(PORTAL.ROCKETCOLLIDERS); i += 1) {
        if (!solid) {
            ds_map_add(originalX, id, x);
            x += leftmostX - bbox_right - 10;
        }
    }

    for (j = 0; j < max; j += 1) {
        distance = execute_string(PORTAL.Raycast, PORTAL.ROCKETCOLLIDERS, 2000, x1, y1, x1 + lengthdir_x(2000, dir), y1 + lengthdir_y(2000, dir));
        if (distance > 1) {
            ds_list_add(points, x1);
            ds_list_add(points, y1);
            ds_list_add(points, distance);
            ds_list_add(points, dir);
        }

        new = false;
        if (hitInstance != noone) {
            if (hitInstance.object_index == PORTAL.Portal) {
                if (instance_exists(hitInstance.pair) == true) {
                    dir = (180 + hitInstance.image_angle - dir + hitInstance.pair.image_angle) mod 360;
                    x1 = hitInstance.pair.x;
                    y1 = hitInstance.pair.y;

                    // Remove the portal so we dont collide with it again.
                    with(hitInstance.pair) {
                        ds_map_add(originalX, id, x);
                        x += leftmostX - bbox_right - 10;
                    }

                    new = true;
                }
            }
        }

        if (new == false) {
            break;
        }
    }

    var key, origX;
    for(key = ds_map_find_first(originalX); key; key = ds_map_find_next(originalX, key)) {
        origX = ds_map_find_value(originalX, key);
        with(key) {
            x = origX;
        }
    }

    ds_map_destroy(originalX);
    gunUnsetSolids();
');

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PORTAL.CubeSpawner = object_add();
object_set_parent(PORTAL.CubeSpawner, PORTAL.Receiver);
object_set_sprite(PORTAL.CubeSpawner, -1);

object_event_add(PORTAL.CubeSpawner, ev_create, 0, '
    event_inherited();
    wasActive = false;
    auto = false;
    ejectSpeed = 0;
    form = 0;

    cube = instance_create(-999, -999, PORTAL.Cube);
    cube.image_alpha = 0;
');

object_event_add(PORTAL.CubeSpawner, ev_step, ev_step_normal, '
    event_inherited();

    if ((auto == false && active != wasActive && active == true) || (auto == true && cube.image_alpha == 0 && cube.held == false)) {
        if (cube.image_alpha == 1) {
            var fizzle;
            fizzle = instance_create(cube.x, cube.y, PORTAL.FizzleCube);
            with(fizzle) {
                sprite_index = other.cube.sprite_index;
                image_index = other.cube.image_index;
                image_xscale = other.cube.image_xscale;
                motion_add(other.cube.direction, other.cube.speed);
            }
        }

        cube.direction = image_angle;
        cube.speed = ejectSpeed;
        cube.x = x;
        cube.y = y;
        cube.image_alpha = 1;
        cube.image_index = form;
        cube.held = false;
    }

    wasActive = active;
');

object_event_add(PORTAL.CubeSpawner, ev_other, ev_user1, '
    event_inherited();
    auto = readProperty(other.properties, "auto", BOOL, auto);
    ejectSpeed = readProperty(other.properties, "speed", REAL, ejectSpeed);
    form = readProperty(other.properties, "form", REAL, form);
');

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PORTAL.Elevator = object_add();
object_set_sprite(PORTAL.Elevator, sprite64);
object_set_visible(PORTAL.Elevator, false);

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PORTAL.Notification = object_add();
object_set_sprite(PORTAL.Notification, sprite64);
object_set_visible(PORTAL.Notification, false);

object_event_add(PORTAL.Notification, ev_create, 0, '
    notified = false;
    message = "The default message";
');

object_event_add(PORTAL.Notification, ev_other, ev_user1, '
    message = readProperty(other.properties, "message", STRING, message);
');
